{************************************}
{      Copyright 2002 Cyrille2       }
{        Tous  droits rservs       }
{ -------- Cyrille2@free.fr -------- }
{ --- http://www.visiomaster.net --- }
{************************************}

unit Puissance4;

interface

uses
  Classes, Graphics, Controls,
  ExtCtrls, Sysutils;

type
  TPlayer = 1..2;

  TEndMatchType = (MathNull,Joueur1,Joueur2);
  TEndMatch = Procedure(Sender : TObject; MatchType : TEndMatchType) Of object;

  TP4ErrorType = (General,InvalidParam,NotRound,FullCol,GameOff);
  TP4Error = Procedure(Sender : TObject; P4Error : TP4ErrorType; Player : TPlayer) Of object;

  TClickGrille = Procedure(Sender : TObject; Button: TMouseButton; Col : Integer) Of object;

  TGameInfo = Procedure(Sender : TObject; Player : TPlayer; Col, Row : Integer) Of object;

  TPuissance4 = class(TcustomPanel)
  private
    { Private declarations }
    FEndMatch : TEndMatch;
    FP4Error : TP4Error;
    FGameInfo : TGameInfo;
    FClickGrille : TClickGrille;

    FShape : TShapeType;
    FPlayersColor : Array[1..2] of TColor;
    procedure SetPlaying(Value : Boolean);
    procedure SetShape(Value : TShapeType);
    procedure SetColorPlayer1(Value : TColor);
    procedure SetColorPlayer2(Value : TColor);

  public
    { Public declarations }
   // Data : Pointer;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Restart(FirstPlayer : TPlayer);
    procedure Play(Col : Integer; Joueur : TPlayer);

  protected
    FPlaying : Boolean;
    FGrille : Array[1..6,1..6] of TShape;
    FNextPlayer : TPlayer;

    procedure GrilleMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    function GetPlayerByColor(TheColor : TColor) : Integer;
    procedure CheckWinner(Colonne, Ligne : Byte);
    procedure AddPions(Colonne : Byte; Joueur : TPlayer);
    procedure Resize; override;

  published          
    property Align;
    property Anchors;
    property Player1 : TColor Read FPlayersColor[1] write SetColorPlayer1 default clGreen;
    property Player2 : TColor Read FPlayersColor[2] write SetColorPlayer2 default clRed;
    property Playing : Boolean Read  FPlaying write SetPlaying default False;
    property FirstPlayer : TPlayer read FNextPlayer write FNextPlayer default 1;
    property PopupMenu;
    property Shape : TShapeType read FShape write SetShape default stCircle;
    property Visible;

    property OnClickGrille : TClickGrille read FClickGrille write FClickGrille;
    property OnPlay : TGameInfo read FGameInfo write FGameInfo;
    property OnEndMatch : TEndMatch read FEndMatch write FEndMatch;
    property OnError : TP4Error read FP4Error write FP4Error;
  end;

procedure Register;

implementation

uses Variants;

procedure Register;
begin
  RegisterComponents('Cyrille2', [TPuissance4]);
end;

constructor TPuissance4.Create(AOwner: TComponent);
var X, Y : Integer;
begin
    inherited Create(AOwner);

    SetBounds(0,0,401,346);

    FPlayersColor[1] := clGreen;
    FPlayersColor[2] := clRed;
    FNextPlayer := 1;
    FPlaying := False;
    FShape := stCircle;

    for Y := 1 To 6 Do
      for X := 1 To 7 Do
      begin
          FGrille[X][Y] := TShape.Create(Self);
          FGrille[X][Y].Shape := FShape;

          FGrille[X][Y].Parent := Self;
          FGrille[X][Y].Tag := X;

          FGrille[X][Y].OnMouseDown := GrilleMouseDown;
      end;
end;

destructor TPuissance4.Destroy;    
var X, Y : Integer;
begin

    for Y := 1 To 6 Do
      for X := 1 To 6 Do
      begin
          FGrille[X][Y] := TShape.Create(Self);
          FGrille[X][Y].Shape := FShape;

          FGrille[X][Y].Parent := Self;
          FGrille[X][Y].Tag := X;

          FGrille[X][Y].OnMouseDown := GrilleMouseDown;
      end;

  inherited Destroy;
end;

procedure TPuissance4.Resize;
var X, Y : Byte;
    ShapeSize : Integer;
begin
    Self.Height := 86 * Self.Width div 100;
    ShapeSize := 12 * Self.Width div 100;

    for Y := 1 To 6 Do
        for X := 1 To 9 Do
            FGrille[X,Y].free;
end;

function TPuissance4.GetPlayerByColor(TheColor : TColor) : Integer;
var I :integer;
begin
    Result := -1;

    if FPlayersColor[1] = TheColor Then
        Result := 1
    else
        Result := 2;
end;

procedure TPuissance4.CheckWinner(Colonne, Ligne : Byte);
var NbRestants, X, Y : Byte;
    FWinColor : TColor;

begin
    FWinColor := clWhite;
    NbRestants := 0;

    for Y := 1 To 6 Do
        for X := 1 To 6 Do
            if FGrille[X][Y].Brush.Color = clWhite Then
                Inc(NbRestants)
            else
              if (
                    (X <= 4) and
                    (
                      (
                        (FGrille[X+1][Y].Brush.Color = FGrille[X][Y].Brush.Color) and
                        (FGrille[X+2][Y].Brush.Color = FGrille[X][Y].Brush.Color) and
                        (FGrille[X+3][Y].Brush.Color = FGrille[X][Y].Brush.Color)
                      )
                      or  // horizontal // pench bas gauche, haut droit
                      (
                        (Y <= 3) and
                        (
                          (FGrille[X+1][Y+1].Brush.Color = FGrille[X][Y].Brush.Color) and
                          (FGrille[X+2][Y+2].Brush.Color = FGrille[X][Y].Brush.Color) and
                          (FGrille[X+3][Y+3].Brush.Color = FGrille[X][Y].Brush.Color)
                        )
                      )
                    )
                 )

                 or

                 (
                    (X >= 4) and
                    (Y <= 3) and
                    (
                      (FGrille[X-1][Y+1].Brush.Color = FGrille[X][Y].Brush.Color) and
                      (FGrille[X-2][Y+2].Brush.Color = FGrille[X][Y].Brush.Color) and
                      (FGrille[X-3][Y+3].Brush.Color = FGrille[X][Y].Brush.Color)
                    )
                 )

                 or

                 (
                    (Y <= 3) and
                    (
                      (FGrille[X][Y+1].Brush.Color = FGrille[X][Y].Brush.Color) and
                      (FGrille[X][Y+2].Brush.Color = FGrille[X][Y].Brush.Color) and
                      (FGrille[X][Y+3].Brush.Color = FGrille[X][Y].Brush.Color)
                    )
                 )
              Then
            begin
                  FWinColor := FGrille[Colonne][Ligne].Brush.Color;
                  Break;
            end;


    if not (FWinColor = ClWhite) Then
        case GetPlayerByColor(FWinColor) of
           1 :
              begin
                FPlaying := False;

                if Assigned(FEndMatch) Then
                    FEndMatch(Self,Joueur1);
              end;
           2 :
              begin
                FPlaying := False;

                if Assigned(FEndMatch) Then
                    FEndMatch(Self,Joueur2);
              end;
        end
    else
      if NbRestants = 0 Then
      begin
        FPlaying := False;

        if Assigned(FEndMatch) Then
            FEndMatch(Self,MathNull);
      end;
end;

procedure TPuissance4.AddPions(Colonne : Byte; Joueur : TPlayer);
var Y : Integer;
begin
    if not FPlaying Then
    begin
        if Assigned(FP4Error) Then
            FP4Error(Self,GameOff,Joueur);

        Exit
    end
    else
        if not (Joueur = FNextPlayer) Then
        begin
            if Assigned(FP4Error) Then
                FP4Error(Self,NotRound,Joueur);

            Exit;
        end;

    for Y := 1 To 6 Do
        if FGrille[Colonne][Y].Brush.Color = clWhite Then
        begin
            FGrille[Colonne][Y].Brush.Color := FPlayersColor[Joueur];

            if Joueur = 1 Then
                FNextPlayer := 2
            else
                FNextPlayer := 1;

            if Assigned(FGameInfo) Then
                FGameInfo(Self,Joueur,Colonne,Y);


            CheckWinner(Colonne,Y);

            Exit;
        end;

    // s'il n'est pas en arret ici, alors la colonne est pleine

    
    if Assigned(FP4Error) Then
        FP4Error(Self,FullCol, Joueur);
end;

procedure TPuissance4.GrilleMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
    if Assigned(FClickGrille) Then
        FClickGrille(Self,Button,TComponent(Sender).Tag);
end;

procedure TPuissance4.Restart(FirstPlayer : TPlayer);
begin
    FNextPlayer := FirstPlayer;
    SetPlaying(True);
end;

procedure TPuissance4.Play(Col : Integer; Joueur : TPlayer);
begin
    if (Col > 0) and (Col <= 7) Then
        AddPions(Col,Joueur)
    else
        if Assigned(FP4Error) Then
            FP4Error(Self,InvalidParam, Joueur);
end;


{les sets}
procedure TPuissance4.SetPlaying(Value : Boolean);
var X, Y : Byte;
begin
    if Value Then
        For Y := 1 To 6 Do
          For X := 1 To 7 Do
            FGrille[X][Y].Brush.Color := clWhite;

    FPlaying := Value;
end;

procedure TPuissance4.SetShape(Value : TShapeType);
var X, Y : Byte;
begin
    if not (Value = FShape) Then
    begin
        for Y := 1 To 6 Do
            for X := 1 To 7 Do
                FGrille[X][Y].Shape := Value;

        FShape := Value;
    end;
end;

procedure TPuissance4.SetColorPlayer1(Value: TColor);
var X, Y : Byte;
begin
    if not (Value = FPlayersColor[1]) and not (Value = clWhite) Then
    begin
        for Y := 1 To 6 Do
            for X := 1 To 7 Do
                if FGrille[X][Y].Brush.Color = FplayersColor[1] Then
                    FGrille[X][Y].Brush.Color := Value;

        FPlayersColor[1] := Value;
    end;
end;

procedure TPuissance4.SetColorPlayer2(Value: TColor);
var X, Y : Byte;
begin
    if not (Value = FPlayersColor[2]) and not (Value = clWhite) Then
    begin
        for Y := 1 To 6 Do
            for X := 1 To 7 Do
                if FGrille[X][Y].Brush.Color = FplayersColor[2] Then
                    FGrille[X][Y].Brush.Color := Value;

        FPlayersColor[2] := Value;
    end;
end;

end.
